from collections import OrderedDict
from typing import Callable, List, Any, Dict, Union

from algo_ops.ops.op import Op


class Pipeline(Op):
    """
    A generic pipeline is list of operations that execute serially on an input.
    The output of the previous pipeline step is the input of the next pipeline step.
    """

    @staticmethod
    def _pipeline_op_name(op: Op) -> str:
        return str(op) + ":" + str(op.name)

    def _run(self, inp: Any) -> Any:
        """
        Run entire pipeline on input, one op at a time.

        param inp: The top-level pipeline input

        return:
            output: The output of the pipeline
        """
        current_input = inp
        for op_name in self.ops.keys():
            op = self.ops[op_name]
            current_input = op.exec(inp=current_input)
        return current_input

    def __init__(self, ops: List[Op]):
        super().__init__(func=self._run)
        self.ops = OrderedDict()
        for i, op in enumerate(ops):
            assert isinstance(op, Op)
            self.ops[self._pipeline_op_name(op=op)] = op

    @classmethod
    def init_from_funcs(
        cls, funcs: List[Callable], op_class: Union[Any, List[Any]]
    ) -> "Pipeline":
        """
        param funcs: List of pipeline functions that execute serially
            as operations in pipeline.
        param op_class: The subclass of Op that the pipeline uses
        """
        if not isinstance(op_class, list):
            op_class = [op_class for _ in range(len(funcs))]
        ops: List[Op] = list()
        for i, func in enumerate(funcs):
            ops.append(op_class[i](func))
        return cls(ops=ops)

    def set_params(self, params: Dict[str, Any]) -> None:
        raise ValueError(
            "Please use set_pipeline_params when setting params of pipeline."
        )

    def _find_op(self, func_name: str) -> Op:
        """
        Helper function to find an Op corresponding to a pipeline function.

        param func_name: Name of function

        return:
            Found Op (or ValueError if no Op found)
        """
        for key in self.ops.keys():
            op = self.ops[key]
            assert isinstance(op, Op)
            if op.name == func_name:
                return op
        raise ValueError("Op not found: " + func_name)

    def set_pipeline_params(self, func_name: str, params: Dict[str, Any]) -> None:
        """
        Fixes parameters of a function in the pipeline.

        param func_name: The name of the function
        param params: The parameters to fix
        """
        op = self._find_op(func_name=func_name)
        op.set_params(params=params)

    def save_input(self, out_path: str = ".") -> None:
        raise ValueError("Please use save_output to visualize pipeline data flow.")

    def vis_input(self) -> None:
        raise ValueError("Please use vis to visualize pipeline data flow.")

    def vis(self) -> None:
        """
        Visualize current outputs of each Op.
        """
        for i, op_name in enumerate(self.ops.keys()):
            op = self.ops[op_name]
            assert isinstance(op, Op)
            if i == 0:
                op.vis_input()
            op.vis()

    def save_output(self, out_path: str = ".") -> None:
        """
        Saves pipeline Op outputs to file.

        param out_path: Path to where output should go
        """
        for i, op_name in enumerate(self.ops.keys()):
            op = self.ops[op_name]
            assert isinstance(op, Op)
            if i == 0:
                op.save_input(out_path=out_path)
            op.save_output(out_path=out_path)

    def vis_profile(self) -> None:
        """
        Visualizes timing profiling information about pipeline Ops.
        """
        print("---Profile---")
        for i, op_name in enumerate(self.ops.keys()):
            op = self.ops[op_name]
            assert isinstance(op, Op)
            op.vis_profile()
        print(
            "Total: "
            + self._format_execution_time_stats(
                execution_times=list(self.execution_times)
            )
        )
        print("-------------")



import multiprocessing
from typing import Callable, List, Any, Sequence


def _sequential(
    func: Callable, params: Sequence[Any], debug: bool = False
) -> List[Any]:
    """
    Sequentially runs a function.

    param func: The function to run
    param params: The parameters to the function
    param debug: Whether to enable debug prints

    return:
        Results of function executions
    """
    results: List[Any] = list()
    for param in params:
        if debug:
            print(param)
        result = func(param)
        results.append(result)
    return results


def _pool(func: Callable, params: Sequence[Any]) -> List[Any]:
    """
    Implements paraloop using multiprocessing pool.

    param func: The function to parallelize
    param params: Inputs to function
    return:
        Results of function executions
    """
    num_cores = multiprocessing.cpu_count()
    with multiprocessing.Pool(num_cores) as p:
        return p.map(func, params)


def loop(func: Callable, params: Sequence[Any], mechanism: str = "pool") -> List[Any]:
    """
    Executes function calls with list of parameters using specified mechanism.

    param func: The function
    param params: The parameters to use
    param mechanism: The mechanism
    return:
        Results of function executions
    """
    if mechanism == "pool":
        return _pool(func=func, params=params)
    elif mechanism == "sequential":
        return _sequential(func=func, params=params, debug=False)
    elif mechanism == "debug":
        return _sequential(func=func, params=params, debug=True)
    else:
        raise ValueError("Unsupported mechanism: " + str(mechanism))



import os
from typing import Union

import cv2
import numpy as np
from matplotlib import pyplot as plt

from algo_ops.ops.op import Op

"""
CVOps is infrastructure to build an OpenCV pipeline as a list of
feed-forward ops that support op-level visualization tools and debugging.
"""


class CVOp(Op):
    """
    Represents a single computer vision operation that can be executed.
    Inputs and outputs can be visualized.
    """

    @staticmethod
    def _pyplot_image(img: np.array, title: str) -> None:
        """
        Helper function to plot image using pyplot.

        param img: Image to plot
        param title: Image title
        """
        plt.imshow(img)
        plt.title(title)

    def vis_input(self) -> None:
        """
        Plot current input image using pyplot (jupyter compatible)
        """
        self._pyplot_image(img=self.input, title=self.name)

    def vis(self) -> None:
        """
        Plot current output image using pyplot (jupyter compatible)
        """
        self._pyplot_image(img=self.output, title=self.name)

    def save_input(self, out_path: str = ".") -> None:
        """
        Saves current input image to file.

        param out_path: Path to where input image should be saved.
        """
        if self.output is not None:
            outfile = os.path.join(out_path, self.name + "_input.png")
            cv2.imwrite(outfile, self.input)
        else:
            raise ValueError("Op " + str(self.name) + " has not executed yet.")

    def save_output(self, out_path: str = ".") -> None:
        """
        Saves current output image to file.

        param out_path: Path to where output image should be saved.
        """
        if self.output is not None:
            outfile = os.path.join(out_path, self.name + ".png")
            cv2.imwrite(outfile, self.output)
        else:
            raise ValueError("Op " + str(self.name) + " has not executed yet.")

    def exec(self, inp: Union[np.array, str]) -> np.array:
        """
        A CV op takes in either the file name of an image or an image,
        performs an operation on the image, and returns as new image.

        param inp: The input
        return
            output: The result of the operation
        """
        if isinstance(inp, str):
            inp = cv2.imread(filename=inp)
            inp = cv2.cvtColor(inp, cv2.COLOR_BGR2RGB)
        if not isinstance(inp, np.ndarray):
            raise ValueError("Unsupported Input: " + str(inp))
        return super().exec(inp=inp)



import math
from typing import List, Callable

from matplotlib import pyplot as plt

from algo_ops.ops.cv import CVOp
from algo_ops.ops.op import Op
from algo_ops.pipeline.pipeline import Pipeline


class CVPipeline(Pipeline):
    """
    Implementation of an OpenCV Image Processing pipeline in the algo_ops framework that allows
    auto-dashboarding of pipeline steps.
    """

    @classmethod
    def init_from_funcs(cls, funcs: List[Callable], op_class=CVOp) -> "CVPipeline":
        """
        param funcs: List of pipeline functions that execute serially
            as operations in pipeline.
        param op_class: The subclass of Op that the pipeline uses
        """
        assert op_class is CVOp, "Cannot use non-CVOp in CVPipeline."
        op_class = [op_class for _ in range(len(funcs))]
        ops: List[Op] = list()
        for i, func in enumerate(funcs):
            ops.append(op_class[i](func))
        return cls(ops=ops)

    def vis(
        self, num_cols: int = 4, fig_width: int = 15, fig_height: int = 6, dpi: int = 80
    ) -> None:
        """
        Plot current output images of each Op using pyplot (jupyter compatible).
        Defaults optimize for Jupyter notebook plotting.

        param num_cols: Number of image columns to display
        param fig_width: Total width of figure
        param fig_height: Total height of figure
        param dpi: DPI of figure
        """
        num_rows = math.ceil((len(self.ops.keys()) + 1) / num_cols)
        plt.figure(figsize=(fig_width, fig_height), dpi=dpi)
        plt_num = 1
        for i, op_name in enumerate(self.ops.keys()):
            op = self.ops[op_name]
            assert isinstance(op, Op)
            if i == 0:
                plt.subplot(num_rows, num_cols, plt_num)
                op.vis_input()
                plt.title("Input")
                plt_num += 1
            plt.subplot(num_rows, num_cols, plt_num)
            plt_num += 1
            op.vis()
        plt.show()



import collections
import functools
import os
import time
from abc import ABC, abstractmethod
from typing import Callable, List, Any, Dict, Sequence, Optional, Tuple

import numpy as np

import algo_ops.paraloop.paraloop as paraloop
from algo_ops.pickleable_object.pickleable_object import PickleableObject


class Op(ABC, PickleableObject):
    """
    Represents a single algorithm operation that can be executed. Inputs and outputs can be visualized.
    """

    def __init__(self, func: Callable):
        """
        param func: The operation function
        """

        # core functionality
        self.func = func
        self.exec_func = func
        self.name = func.__name__
        self.input = None
        self.output = None

        # profiling
        self.execution_times: collections.deque = collections.deque(maxlen=1000)

        # evaluation functionality variables
        self.eval_func = None
        self.incorrect_pkl_path: Optional[str] = None

    def exec(self, inp: Any) -> Any:
        """
        Executes operation function on an input. Is also self-time profiling.

        param inp: The input
        return
            output: The result of the operation
        """
        self.input = inp
        t0 = time.time()
        self.output = self.exec_func(inp)
        tf = time.time()
        elapsed_time = tf - t0
        self.execution_times.append(elapsed_time)
        return self.output

    @abstractmethod
    def vis_input(self) -> None:
        """
        Visualize current input.
        """
        pass

    @abstractmethod
    def vis(self) -> None:
        """
        Visualize current output.
        """
        pass

    @abstractmethod
    def save_input(self, out_path: str) -> None:
        """
        Saves current input to file.

        param out_path: Path to where input should be saved.
        """
        pass

    @abstractmethod
    def save_output(self, out_path) -> None:
        """
        Saves current output to file.

        param out_path: Path to where output should be saved.
        """
        pass

    def set_params(self, params: Dict[str, Any]) -> None:
        """
        Sets parameters of operation.

        param params: Dict that maps parameter name -> parameter value
        """
        self.exec_func = functools.partial(self.func, **params)

    @staticmethod
    def _format_execution_time_stats(
        execution_times: List[float], num_sf: int = 9
    ) -> str:
        """
        Formats execution time stats.

        param execution_times: List of execution times
        num_sf: The number of significant figures to display

        return
            output: (mean) +/- (std) s/calls
        """
        mean_val = np.mean(execution_times)
        std_val = np.std(execution_times)
        return (
            str(np.round(mean_val, num_sf))
            + " +/- "
            + str(np.round(std_val, num_sf))
            + " s/call"
        )

    def vis_profile(self) -> None:
        """
        Prints execution time statistics of Op.
        """
        print(
            self.name
            + ": "
            + self._format_execution_time_stats(
                execution_times=list(self.execution_times)
            )
        )

    def _embedded_eval(self, inp: Any) -> Tuple[Any, bool]:
        """
        Helper function to embed evaluation and prediction in same function.
        Returns true if function, when run on input, yielded correct result.
        If false, pickles, op state to file, if pickle path is specified.
        """
        result = self.exec(inp=inp)
        correct = self.eval_func(inp=inp, pred=result)
        if isinstance(inp, str):
            inp = inp.replace("/", "_")
        if not correct and self.incorrect_pkl_path is not None:
            outfile = os.path.join(self.incorrect_pkl_path, str(inp) + ".pkl")
            self.to_pickle(out_pkl_path=outfile)
        return result, correct

    def evaluate(
        self,
        inputs: Sequence[Any],
        eval_func: Callable,
        incorrect_pkl_path: Optional[str] = None,
        mechanism: str = "pool",
    ) -> List[Tuple[Any, bool]]:
        """
        Evaluates Op on a set of inputs. On each prediction, an evaluation function is run.
        If the answer is incorrect, the op state is pickled to [incorrect_pkl_path] / [inp].pkl.
        This is useful for debugging and evaluating an Op.

        param inputs: The set of inputs to evaluate on
        param eval_func: A function to evaluate a prediction on an input
        param incorrect_pkl_path: Path where incorrect prediction states should be pickled
        param mechanism: The paraloop mechanism to use

        return:
            results: (List of results of pipeline executions, bool whether answer was correct)
        """
        if incorrect_pkl_path is not None:
            os.makedirs(incorrect_pkl_path, exist_ok=True)
        self.incorrect_pkl_path = incorrect_pkl_path
        self.eval_func = eval_func
        return paraloop.loop(
            func=self._embedded_eval, params=inputs, mechanism=mechanism
        )



import os
import pickle


class PickleableObject:

    def to_pickle(self, out_pkl_path: str) -> None:
        """
        Pickles object.

        param out_pkl_path: Path to output pickle file.
        """
        with open(out_pkl_path, 'wb') as fout:
            pickle.dump(self, fout)
        assert os.path.exists(out_pkl_path)

    @classmethod
    def load_from_pickle(cls, pkl_path: str) -> "PickleableObject":
        """
        Loads object from pickled file.

        param pkl_path: Path to pkl file containing object

        Return:
            loaded_reference: The loaded object
        """
        with open(pkl_path, 'rb') as fin:
            loaded_obj = pickle.load(fin)
        assert isinstance(loaded_obj, cls)
        return loaded_obj



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
with open("requirements.txt") as f:
    required = f.read().splitlines()

setuptools.setup(
    name="algo-ops",
    version="0.0.1.5.1",
    author="Prateek Tandon",
    author_email="prateek1.tandon@gmail.com",
    description="Algo-Ops Smart Infrastructure",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/prateekt/algo-ops",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9.7",
    install_requires=required,
)



from algo_ops.ops.op import Op


class TextOp(Op):
    """
    Represents a single text processing operation that can be executed.
    Inputs and outputs can be printed to screen and saved.
    """

    def vis_input(self) -> None:
        """
        Print current input.
        """
        print("Input: " + str(self.input))

    def vis(self) -> None:
        """
        Print current output.
        """
        print(self.name + ": " + str(self.output))

    def save_input(self, out_path: str) -> None:
        """
        Saves current input text to file.

        param out_path: Path to where input file should be saved.
        """
        if self.input is not None:
            with open(out_path, "w") as out_file:
                out_file.write(self.input)
        else:
            raise ValueError("Op " + str(self.name) + " has not executed yet.")

    def save_output(self, out_path: str) -> None:
        """
        Saves current output to file.

        param out_path: Path to where output file should be saved.
        """
        if self.output is not None:
            with open(out_path, "w") as out_file:
                out_file.write(self.output)
        else:
            raise ValueError("Op " + str(self.name) + " has not executed yet.")
